
 -- This q-object allows creation of a virtually infinite square,
 -- of the specified color, with the rotational "center" at the
 -- specified distance above or below the viewer. It can then be
 -- rotated or moved at this center for a variety of effects.
 -- The polygons created aren't really infinite, but are sized
 -- sufficiently that they give a realistic horizon appearance.
 -- Also, what they're emulating isn't really an infinite surface
 -- (which would look rather bizarre standing on it in real life),
 -- but rather a surface with a drop off at the horizon (which is
 -- close enough to a curved planetary surface.) While the
 -- y-axis rotation, x-movement, and z-movement are implemented,
 -- in reality they all do nothing to the view; the sole purpose
 -- of those functions is to allow any ground texture to be
 -- properly manipulated whenever textures are implemented.

include quartz.e
include qvertex.e
include qpolygon.e

 ----------

constant TRUE = 1

 ----------

without warning

 -- Param contains the color (Param[1]) & the y-coordinate (Param[2])
global function Floor_Floor (qobject Self, object Param)

   qobject vertex1, vertex2, vertex3, vertex4
   atom halfside, x, y

    -- first thing to do, is determine the size of the square to
    -- represent the floor; it must be large enough to look like it
    -- reaches to the horizon even at the minimum y-coord of an
    -- object (including a floor)

    -- the x-value of the square's edge must be at least such that
    -- even at the furthest point, it still translates to a 2D
    -- coordinate at the edge of the screen (or beyond):
   x = MID_X * (Z_HORIZON / Z_DISTANCE)

    -- to make sure we don't get flaky spots at the corners whenever
    -- this square is rotated, we assume that the distance from the
    -- origin to (x, 0, horizon) is half of the length of the square;
    -- distance^2=x^2 + horizon^2; distance=sqrt(x^2+horizon^2):
   halfside = floor (sqrt ((x * x) + (Z_HORIZON * Z_HORIZON)))

    -- now we know the 4 beginning points...
   y = Param[2]
   vertex1 = new_qobject ("Vertex", {-halfside, y,  halfside})
   vertex2 = new_qobject ("Vertex", { halfside, y,  halfside})
   vertex3 = new_qobject ("Vertex", {-halfside, y, -halfside})
   vertex4 = new_qobject ("Vertex", { halfside, y, -halfside})

   Self = set_qmember (Self, "Vertices",
                       {vertex1, vertex2, vertex3, vertex4})
   Self = set_qmember (Self, "Color", Param[1])

   return Self

end function -- end Floor_Floor()


 -- Param is ignored
global function Floor_DrawSelf (qobject Self, object Param)

   sequence Points, Vertices
   atom     Color, z
   qobject  CurrVertex

   Vertices = get_qmember (Self, "Vertices")

   Points = repeat ({}, 4)
   for i = 1 to 4 do
      CurrVertex = Vertices[i]
      z = get_qmember (CurrVertex, "z")
      if (z < MIN_Z) then
          -- move the vertex; this is a work around for now, since
          -- the floor image won't be accurate; however, with the
          -- addition of textures, we'll be plotting each individual
          -- pixel anyway, and we'll only be doing the ones on-screen,
          -- so it won't matter.
         z = MIN_Z
         CurrVertex = set_qmember (CurrVertex, "z", z)
      end if
      Points[i] = send_qmessage(CurrVertex, "Translate", 0)
   end for

   Color = get_qmember (Self, "Color")

   polygon (Color, TRUE, Points)

   return 0

end function -- end Floor_DrawSelf()


 -- currently a do-nothing; when textures are implemented, shift the
 -- texture along the x axis
global function Floor_MoveX (qobject Self, object Param)

   return Self

end function -- end Floor_MoveX()


global function Floor_MoveY (qobject Self, object Param)

   sequence Vertices

   Vertices = get_qmember (Self, "Vertices")

   for i = 1 to 4 do
      Vertices[i] = send_qmessage (Vertices[i], "MoveY", Param)
   end for

   Self = set_qmember (Self, "Vertices", Vertices)

   return Self

end function -- end Floor_MoveY()


 -- currently a do-nothing; when textures are implemented, shift the
 -- texture along the z axis
global function Floor_MoveZ (qobject Self, object Param)

   return Self

end function -- end Floor_MoveZ()


 -- like the Rotate method for a Vertex, this rotates around a
 -- specified central point included with the degree specifications;
 -- if there's no specified center, the origin is assumed.
global function Floor_Rotate (qobject Self, object Param)

   sequence Vertices

   Vertices = get_qmember (Self, "Vertices")

   for i = 1 to 4 do
      Vertices[i] = send_qmessage (Vertices[i], "Rotate", Param)
   end for

   Self = set_qmember (Self, "Vertices", Vertices)

   return Self

end function -- end Floor_Rotate()

with warning

 ----------

define_qclass ("Floor", BASE_QCLASS,
               {"Color", "Vertices"},
               {"DrawSelf", "MoveX", "MoveY", "MoveZ", "Rotate"},
               {routine_id ("Floor_DrawSelf"),
                routine_id ("Floor_MoveX"),
                routine_id ("Floor_MoveY"),
                routine_id ("Floor_MoveZ"),
                routine_id ("Floor_Rotate")},
               routine_id ("Floor_Floor"))

