include quartz.e
include graphics.e
include qvertex.e
include qpolygon.e

 ----------

constant DEFAULT_VPOLYGON = {0, 0, 0}

 ----------

 -- treats the given sequence like a mathematical set; the given
 -- object is only added if it is not already in the sequence
function AddToSet (sequence s, object x)

   if (find (x, s) != 0) then
      return s
   else
      return prepend (s, x)
   end if

end function

 ----------

 ----------

without warning

 -- the parameter is a list of polygons (q-objects)
global function PolyObject_PolyObject (qobject Self, object Param)

   integer  L
   sequence Vertices, VertexSet, VPolygons
   qobject  CurrPolygon
   object   temp, CurrVPolygon

   L = length (Param)

    -- to begin, we make things easier by extracting all of the
    -- actual vertices from the polygons & avoiding redundant calls
    -- to "get_qmember"
   Vertices = repeat ({}, L)
   for i = 1 to L do
      CurrPolygon = Param[i]
      Vertices[i] = get_qmember (CurrPolygon, "Vertices")
   end for

    -- first create a set of all polygon vertices, so a vertex won't
    -- be recalculated more than once
   VertexSet = {}
   for i = 1 to L do
      CurrVPolygon = Vertices[i]
      for j = 1 to 3 do
         VertexSet = AddToSet (VertexSet, CurrVPolygon[j])
      end for
   end for
   Self = set_qmember (Self, "VertexSet", VertexSet)

    -- next we map out virtual-polygons (based on the vertices)
    -- so we can map the vertex changes to the polygons
   VPolygons = repeat ({}, L)
   for i = 1 to L do
      temp = Vertices[i]
      CurrVPolygon = DEFAULT_VPOLYGON
      CurrVPolygon[1] = find (temp[1], VertexSet)
      CurrVPolygon[2] = find (temp[2], VertexSet)
      CurrVPolygon[3] = find (temp[3], VertexSet)
      VPolygons[i] = CurrVPolygon
   end for
   Self = set_qmember (Self, "VPolygons", VPolygons)

    -- finally, we compute the center
   Self = set_qmember (Self, "Center",
                       send_qmessage (Self, "FindCenter", 0))

    -- and assign the polygons to their member
   Self = set_qmember (Self, "Polygons", Param)

   return Self

end function -- end PolyObject_PolyObject()


 -- this is a bit more complicated than for a single polygon, since
 -- we will be drawing polygons in order from back to front; eventually
 -- it might be altered to not draw "back" polygons at all.
global function PolyObject_DrawSelf (qobject Self, object Param)

   sequence Polygons
   object   junk

    -- first, we extract the polygons, modifying them based on
    -- the vertex changes
   Polygons = send_qmessage (Self, "GetPolygons", 0)

    -- if we're going to cull any back-facing polygons, do it here...

    -- now we order the polygons so we draw them appropriately;
    -- we'll use a simple z-depth sort, which won't always be
    -- accurate, but in high-motion situations it won't matter
   Polygons = ZSort (Polygons, DESCENDING)

    -- finally, we instruct each polygon to draw itself
   for i = 1 to length (Polygons) do
      junk = send_qmessage (Polygons[i], "DrawSelf", 0)
   end for

   return 0

end function -- end PolyObject_DrawSelf()


global function PolyObject_MoveX (qobject Self, object Param)

   sequence VertexSet
   qobject  Center

   VertexSet = get_qmember (Self, "VertexSet")
   Center   = get_qmember (Self, "Center")

   for i = 1 to length (VertexSet) do
      VertexSet[i] = send_qmessage (VertexSet[i], "MoveX", Param)
   end for

   Center = send_qmessage (Center, "MoveX", Param)

   Self = set_qmember (Self, "VertexSet", VertexSet)
   Self = set_qmember (Self, "Center", Center)

   return Self

end function -- end PolyObject_MoveX()


global function PolyObject_MoveY (qobject Self, object Param)

   sequence VertexSet
   qobject  Center

   VertexSet = get_qmember (Self, "VertexSet")
   Center   = get_qmember (Self, "Center")

   for i = 1 to length (VertexSet) do
      VertexSet[i] = send_qmessage (VertexSet[i], "MoveY", Param)
   end for

   Center = send_qmessage (Center, "MoveY", Param)

   Self = set_qmember (Self, "VertexSet", VertexSet)
   Self = set_qmember (Self, "Center", Center)

   return Self

end function -- end PolyObject_MoveY()


global function PolyObject_MoveZ (qobject Self, object Param)

   sequence VertexSet
   qobject  Center

   VertexSet = get_qmember (Self, "VertexSet")
   Center   = get_qmember (Self, "Center")

   for i = 1 to length (VertexSet) do
      VertexSet[i] = send_qmessage (VertexSet[i], "MoveZ", Param)
   end for

   Center = send_qmessage (Center, "MoveZ", Param)

   Self = set_qmember (Self, "VertexSet", VertexSet)
   Self = set_qmember (Self, "Center", Center)

   return Self

end function -- end PolyObject_MoveZ()


 -- NOTE: A PolyObject will rotate around the vertex defined as its
 --    center, not necessarily the origin. This is so that a default
 --    rotation is always around the poly-object's center. To rotate
 --    around the origin, or some other point, set the center to be
 --    the desired point, do the rotation, then re-find the center.
 -- THE PARAMETER IS A SEQUENCE OF DEGREES TO ROTATE PER AXIS, I.E.
 -- {x-rotation, y-rotation, z-rotation}
global function PolyObject_Rotate (qobject Self, object Param)

   sequence VertexSet, DegreesAndCenter
   qobject  Center

   VertexSet        = get_qmember (Self, "VertexSet")
   Center           = get_qmember (Self, "Center")

   if (compare (Center, ORIGIN_VERTEX) = 0) then
       -- the center is the origin; no need to pass it
      DegreesAndCenter = Param
   else
      DegreesAndCenter = append (Param,
                                 send_qmessage (Center, "GetSeq", 0))
   end if

   for i = 1 to length (VertexSet) do
      VertexSet[i] = send_qmessage (VertexSet[i], "Rotate",
                                    DegreesAndCenter)
   end for

   Self = set_qmember (Self, "VertexSet", VertexSet)

   return Self

end function -- end PolyObject_Rotate()


 -- NOTE: A PolyObject only scales from it's central point; the
 --    amount of scaling for each dimension is required. Scaling
 --    automatically re-calculates the center, but doesn't set it.
global function PolyObject_Scale (qobject Self, object Param)

   sequence VertexSet
   qobject  Center, Vertex
   atom     x, y, z

   Center = send_qmessage (Self, "FindCenter", 0)

   if (compare (Center, ORIGIN_VERTEX) = 0) then
       -- the center is the origin; no need to shift

      VertexSet = get_qmember (Self, "VertexSet")

      for i = 1 to length (VertexSet) do
         VertexSet[i] = send_qmessage (VertexSet[i], "Scale", Param)
      end for

      Self = set_qmember (Self, "VertexSet", VertexSet)

   else

      x = get_qmember (Center, "x")
      y = get_qmember (Center, "y")
      z = get_qmember (Center, "z")

      Self = send_qmessage (Self, "MoveX", -x)
      Self = send_qmessage (Self, "MoveY", -y)
      Self = send_qmessage (Self, "MoveZ", -z)

      VertexSet = get_qmember (Self, "VertexSet")

      for i = 1 to length (VertexSet) do
         VertexSet[i] = send_qmessage (VertexSet[i], "Scale", Param)
      end for

      Self = set_qmember (Self, "VertexSet", VertexSet)

      Self = send_qmessage (Self, "MoveX", x)
      Self = send_qmessage (Self, "MoveY", y)
      Self = send_qmessage (Self, "MoveZ", z)

   end if

   return Self

end function -- end PolyObject_Scale()


 -- Param is ignored
global function PolyObject_GetPolygons (qobject Self, object Param)

   sequence VertexSet, Polygons, VPolygons, CurrVPolygon
   qobject  CurrPolygon

    -- we extract the polygons, modifying them based on
    -- the vertex changes
   VertexSet = get_qmember (Self, "VertexSet")
   Polygons  = get_qmember (Self, "Polygons")
   VPolygons = get_qmember (Self, "VPolygons")
   for i = 1 to length (Polygons) do
      CurrPolygon  = Polygons[i]
      CurrVPolygon = VPolygons[i]
      CurrPolygon = set_qmember (CurrPolygon, "Vertices",
                       {VertexSet[CurrVPolygon[1]],
                        VertexSet[CurrVPolygon[2]],
                        VertexSet[CurrVPolygon[3]]})
      Polygons[i] = CurrPolygon
   end for

   return Polygons

end function -- end PolyObject_GetPolygons()


 -- Param is ignored
global function PolyObject_FindCenter (qobject Self, object Param)

   integer  NumVertices
   atom     SumX, SumY, SumZ
   sequence VertexSet
   qobject  CurrVertex

   VertexSet   = get_qmember (Self, "VertexSet")
   NumVertices = length (VertexSet)

   SumX = 0
   SumY = 0
   SumZ = 0

   for i = 1 to NumVertices do
      CurrVertex = VertexSet[i]
      SumX = SumX + get_qmember (CurrVertex, "x")
      SumY = SumY + get_qmember (CurrVertex, "y")
      SumZ = SumZ + get_qmember (CurrVertex, "z")
   end for

   return new_qobject ("Vertex", {SumX, SumY, SumZ} / NumVertices)

end function -- end PolyObject_FindCenter()

with warning

 ----------

define_qclass ("PolyObject", BASE_QCLASS,
               {"VertexSet", "VPolygons", "Polygons", "Center"},
               {"DrawSelf", "MoveX", "MoveY", "MoveZ", "Rotate",
                "Scale", "GetPolygons", "FindCenter"},
               {routine_id ("PolyObject_DrawSelf"),
                routine_id ("PolyObject_MoveX"),
                routine_id ("PolyObject_MoveY"),
                routine_id ("PolyObject_MoveZ"),
                routine_id ("PolyObject_Rotate"),
                routine_id ("PolyObject_Scale"),
                routine_id ("PolyObject_GetPolygons"),
                routine_id ("PolyObject_FindCenter")},
               routine_id ("PolyObject_PolyObject"))

