
include quartz.e
include graphics.e

 ----------

global constant G_MODE  = 19 -- 261 is 1024x768 w/ 256 colors
   -- 257 is 640x480 w/ 256 colors

global constant MAX_X  = 320
global constant MAX_Y  = 200
global constant MID_X  = floor (MAX_X / 2)
global constant MID_Y  = floor (MAX_Y / 2)

global constant COLOR_BITS = 8 -- meaning 256 colors available

global constant MAX_VX = 1024 -- arbitrary
global constant MAX_VY = 1024 -- arbitrary
global constant Z_DISTANCE = 256
global constant Z_HORIZON = MAX_VY * Z_DISTANCE + 1
   -- Z_HORIZON is how far away things (up to MAX_VY height) "vanish"
global constant MIN_Z = 10

 ----------

constant PI = 3.141592653589793238

atom     rad
sequence SinTable, CosTable -- these will be in degrees; considering
                            -- that 2pi*radians = 360degrees, we know
                            -- that rad = deg*pi/180 and
                            -- that deg = rad*180/pi

SinTable = repeat (1.0, 360)
CosTable = SinTable

constant PI_FACTOR = PI / 180

for i = 1 to 360 do
   rad = PI_FACTOR * i
   SinTable[i] = sin (rad)
   CosTable[i] = cos (rad)
end for

global constant SIN_TABLE = SinTable
global constant COS_TABLE = CosTable

 ----------

without warning

global function Vertex_Vertex (qobject Self, object Param)

   Self = set_qmember (Self, "x", Param[1])
   Self = set_qmember (Self, "y", Param[2])
   Self = set_qmember (Self, "z", Param[3])

   return Self

end function -- end Vertex_Vertex()


 -- this translates a 3D vertex to it's 2D-perspective point;
 -- this also translates from origin-coordinates to screen-coordinates
global function Vertex_Translate (qobject Self, object Param)

   atom x, y, z, factor

   x = get_qmember (Self, "x")
   y = get_qmember (Self, "y")
   z = get_qmember (Self, "z")

    -- to do this, we map the 3D vertex to a 2D-perspective point;
    -- we distort x and y based on distance (z), for perspective.
    -- The viewer is at z=0; the amount to distort is based on
    -- Z_DISTANCE, the distance the screen is set from the viewer

   factor = Z_DISTANCE /  z

    -- if the x value is at the halfway point, no shift is necessary
   if (x != 0) then
      x = floor (factor * x)
   end if

    -- if the y value is at the halfway point, no shift is necessary
   if (y != 0) then
      y = floor (factor * y)
   end if

   return {MID_X + x, MID_Y - y}

end function -- end Vertex_Translate()


global function Vertex_MoveX (qobject Self, object Param)

   atom x

   x = get_qmember (Self, "x") + Param

   if (x <= MAX_VX) then
      if (x >= -MAX_VX) then
         Self = set_qmember (Self, "x", x)
      end if
   end if

   return Self

end function -- end Vertex_MoveX()


global function Vertex_MoveY (qobject Self, object Param)

   atom y

   y = get_qmember (Self, "y") + Param

   if (y <= MAX_VY) then
      if (y >= -MAX_VY) then
         Self = set_qmember (Self, "y", y)
      end if
   end if

   return Self

end function -- end Vertex_MoveY()


global function Vertex_MoveZ (qobject Self, object Param)

   atom z

   z = get_qmember (Self, "z") + Param

   if (z >= MIN_Z) then
      Self = set_qmember (Self, "z", z)
   end if

   return Self

end function -- end Vertex_MoveZ()


 -- rotation around the specified vertex (a simple sequence);
 -- if no point specified, assumes the origin.
global function Vertex_Rotate (qobject Self, object Param)

   atom     oldx, oldy, oldz,
            newx, newy, newz,
            degx, degy, degz,
            ctrx, ctry, ctrz,
            rad, temp,
            cosrad, sinrad, cosdeg, sindeg
   integer  NotOrigin
   sequence center

   oldx = get_qmember (Self, "x")
   oldy = get_qmember (Self, "y")
   oldz = get_qmember (Self, "z")

   degx = remainder (Param[1], 360)
   degy = remainder (Param[2], 360)
   degz = remainder (Param[3], 360)
   --if (degx = 0) then degx = 360 end if
   --if (degy = 0) then degy = 360 end if
   --if (degz = 0) then degz = 360 end if

   NotOrigin = length (Param) > 3

   if (NotOrigin) then
       -- a point was specified; we won't just be doing things around
       -- the origin...
      center = Param[4]
      ctrx = center[1]
      ctry = center[2]
      ctrz = center[3]
       -- we shift things & act as if we were going around the origin
      oldx = oldx - ctrx
      oldy = oldy - ctry
      oldz = oldz - ctrz
   end if

    -- now to actually rotate the vertex...
    -- first, around the x axis
   if (degx != 0) then
      if (integer (degx) and (degx > 0)) then
         cosdeg = COS_TABLE[degx]
         sindeg = SIN_TABLE[degx]
         newy = (oldy * cosdeg) - (oldz * sindeg)
         newz = (oldz * cosdeg) + (oldy * sindeg)
      else
         rad = PI_FACTOR * degx
         cosrad = cos (rad)
         sinrad = sin (rad)
         newy = (oldy * cosrad) - (oldz * sinrad)
         newz = (oldz * cosrad) + (oldy * sinrad)
      end if
   else
      newy = oldy
      newz = oldz
   end if

    -- now, around the y axis
   if (degy != 0) then
      if (integer (degy) and (degy > 0)) then
         cosdeg = COS_TABLE[degy]
         sindeg = SIN_TABLE[degy]
         newx = (oldx * cosdeg) + (newz * sindeg)
         newz = (newz * cosdeg) - (oldx * sindeg)
      else
         rad = PI_FACTOR * degy
         cosrad = cos (rad)
         sinrad = sin (rad)
         newx = (oldx * cosrad) + (newz * sinrad)
         newz = (newz * cosrad) - (oldx * sinrad)
      end if
   else
      newx = oldx
   end if

    -- finally, around the z axis
   if (degz != 0) then
      temp = newx
      if (integer (degz) and (degz > 0)) then
         cosdeg = COS_TABLE[degz]
         sindeg = SIN_TABLE[degz]
         newx = (temp * cosdeg) - (newy * sindeg)
         newy = (newy * cosdeg) + (temp * sindeg)
      else
         rad = PI_FACTOR * degz
         cosrad = cos (rad)
         sinrad = sin (rad)
         newx = (temp * cosrad) - (newy * sinrad)
         newy = (newy * cosrad) + (temp * sinrad)
      end if
   end if

    -- if we had a non-origin center, shift the vertex back
   if (NotOrigin) then
      newx = newx + ctrx
      newy = newy + ctry
      newz = newz + ctrz
   end if

    -- store the altered coordinates
   Self = set_qmember (Self, "x", newx)
   Self = set_qmember (Self, "y", newy)
   Self = set_qmember (Self, "z", newz)

   return Self

end function -- end Vertex_Rotate()


 -- the Param is a sequence; each element is how much to scale per
 -- dimension: {x-scaling, y-scaling, z-scaling}
 -- NOTE: ALWAYS ASSUMES THE ORIGIN AS THE POINT TO SCALE FROM
global function Vertex_Scale (qobject Self, object Param)

   atom x, y, z

   x = get_qmember (Self, "x") * Param[1]
   y = get_qmember (Self, "y") * Param[2]
   z = get_qmember (Self, "z") * Param[3]

   Self = set_qmember (Self, "x", x)
   Self = set_qmember (Self, "y", y)
   Self = set_qmember (Self, "z", z)

   return Self

end function -- end Vertex_Scale()


global function Vertex_GetSeq (qobject Self, object Param)

   return {get_qmember (Self, "x"),
           get_qmember (Self, "y"),
           get_qmember (Self, "z")}

end function -- end Vertex_GetSeq()

with warning

 ----------

define_qclass ("Vertex", BASE_QCLASS,
               {"x", "y", "z"},
               {"Translate", "MoveX", "MoveY", "MoveZ", "Rotate",
                "Scale", "GetSeq"},
               {routine_id ("Vertex_Translate"),
                routine_id ("Vertex_MoveX"),
                routine_id ("Vertex_MoveY"),
                routine_id ("Vertex_MoveZ"),
                routine_id ("Vertex_Rotate"),
                routine_id ("Vertex_Scale"),
                routine_id ("Vertex_GetSeq")},
               routine_id ("Vertex_Vertex"))

global constant ORIGIN_VERTEX = new_qobject ("Vertex", {0, 0, 0})

